import textwrap
from argparse import RawDescriptionHelpFormatter

from mcmd.commands._registry import arguments
from mcmd.core.command import command
from mcmd.core.compatibility import deprecated
from mcmd.core.errors import McmdError
from mcmd.io import io
from mcmd.io.io import highlight
from mcmd.molgenis import api
from mcmd.molgenis.client import post, put
from mcmd.molgenis.resources import one_resource_exists, ensure_resource_exists, ResourceType
from mcmd.molgenis.security import security


# =========
# Arguments
# =========

@arguments('enable')
def add_arguments(subparsers):
    p_enable = subparsers.add_parser('enable',
                                     help='enable resources and functionality',
                                     description="run 'mcmd enable rls -h' to view the help for those sub-commands")
    p_enable_subparsers = p_enable.add_subparsers(dest="type")

    p_enable_rls = p_enable_subparsers.add_parser('rls',
                                                  help='enables row level security on an entity type')
    p_enable_rls.set_defaults(func=enable_rls,
                              write_to_history=True)
    p_enable_rls.add_argument('entity',
                              type=str,
                              help="the entity type to secure")

    p_enable_theme = p_enable_subparsers.add_parser('theme',
                                                    help='enables a bootstrap theme (deprecated since 8.6)',
                                                    formatter_class=RawDescriptionHelpFormatter,
                                                    description=textwrap.dedent(
                                                        """
                                                        Enable a CSS theme. 

                                                        Deprecated since 8.6. The themes can now be configured in the 
                                                        Application settings. Read https://molgenis.gitbook.io/molgenis/configuration/guide-customize#style-theme
                                                        to learn more. 

                                                        Example usage:
                                                          mcmd enable theme theme3.css
                                                        """
                                                    )
                                                    )
    p_enable_theme.set_defaults(func=enable_theme,
                                write_to_history=True)
    p_enable_theme.add_argument('theme',
                                type=str,
                                help='the bootstrap theme you want to enable, specify the name with or without '
                                     '(.min).css and with or without bootstrap- prefix.')

    p_enable_language = p_enable_subparsers.add_parser('language',
                                                       help='enables a language')
    p_enable_language.set_defaults(func=enable_language,
                                   write_to_history=True)
    p_enable_language.add_argument('language',
                                   type=str,
                                   help="the language you want to enable, specified by the two letter code (e.g. 'en')")


# =======
# Methods
# =======

@command
def enable_rls(args):
    io.start('Enabling row level security on entity type %s' % highlight(args.entity))

    ensure_resource_exists(args.entity, ResourceType.ENTITY_TYPE)
    security.enable_row_level_security(args.entity)


@command
@deprecated(since='8.6.0',
            action='Enabling themes',
            info="Themes can be activated in the app settings now. Use the 'set' command to set the 'legacy_theme_url' "
                 "and 'theme_url'. Read https://molgenis.gitbook.io/molgenis/configuration/guide-customize#style-theme"
                 " to learn more.")
def enable_theme(args):
    """
    enable_theme enables a bootstrap theme
    :param args: commandline arguments containing the id of the theme (without .css)
    :exception McmdError: when applying the theme fails
    :return None
    """
    theme = args.theme.replace('.css', '').replace('.min', '')
    io.start('Applying theme {}'.format(highlight(theme)))
    # Resource can be bootstrap-name.min.css (if molgenis theme), or name.min.css (if uploaded .min.css), or
    # name.css (if uploaded as .css).
    if one_resource_exists([theme + '.min.css', theme + '.css', 'bootstrap-' + theme + '.min.css'], ResourceType.THEME):
        # Molgenis themes start with bootstrap- but this is stripped from the name in the theme manager
        try:
            post(api.set_theme(), data=theme)
        except:
            post(api.set_theme(), data=theme.split('bootstrap-')[1])
    else:
        raise McmdError(
            'Applying theme failed. No themes found containing {} in the name'.format(args.theme, 'sys_set_StyleSheet'))


@command
def enable_language(args):
    io.start('Enabling language {}'.format(highlight(args.language)))
    url = api.rest1('sys_Language/{}/active'.format(args.language))
    put(url, 'true')



import argparse

from mcmd.args._raising_parser import RaisingArgumentParser
from mcmd.args.errors import ArgumentSyntaxError


class ParseKeyValue(argparse.Action):
    """
    Parses key/value pairs, separated by an '=' sign. Returns a dictionary.
    Will parse the values 'true' and 'false' as booleans. As a consequence, string literals 'true' and 'false' are not
    supported.
    """

    def __call__(self, parser: RaisingArgumentParser, namespace, values, option_string=None):
        d = {}

        if values:
            for item in values:
                split_items = item.split("=", 1)

                if len(split_items) != 2:
                    raise ArgumentSyntaxError("Not a valid key/value pair: {}".format(item),
                                              parser.get_usage_as_string())

                key = split_items[0].strip()
                value = ParseKeyValue.__to_bool_or_str(split_items[1])

                d[key] = value

        setattr(namespace, self.dest, d)

    @staticmethod
    def __to_bool_or_str(value: str):
        if value == 'true':
            return True
        elif value == 'false':
            return False
        else:
            return value



"""
On-disk memory. Used to store data between separate commander runs.
"""

import pickle
from datetime import datetime
from typing import Optional

from packaging.version import Version

from mcmd.core.context import context
from mcmd.core.errors import McmdError

_store = {
    'last_version_check': datetime.now(),
    'update_available': None
}


def load():
    try:
        if context().get_storage_file().exists():
            with context().get_storage_file().open('rb') as f:
                _store.update(pickle.load(f))
    except (pickle.PickleError, IOError) as e:
        raise McmdError("Unable to read storage from file: {}".format(e.message))


def get_last_version_check() -> datetime:
    return _store['last_version_check']


def set_last_version_check(last_check: datetime):
    _store['last_version_check'] = last_check
    _persist()


def get_update_available() -> Optional[Version]:
    return _store['update_available']


# noinspection PyTypeChecker
def set_update_available(version: Optional[Version]):
    _store['update_available'] = version
    _persist()


def _persist():
    try:
        with context().get_storage_file().open('wb') as f:
            pickle.dump(_store, f)
    except (pickle.PickleError, IOError) as e:
        raise McmdError("Unable to save storage to file: {}".format(e.message))



"""
A line parser for MCMD scripts. For a full overview of the specs, see the script_spec file in the script package.
"""
from jinja2 import TemplateSyntaxError
from parsy import regex, string, seq, whitespace, from_enum, generate, ParseError

from mcmd.script.model.lines import Line
from mcmd.script.model.statements import Statement, Value, Input, Wait, VisibleComment, Command, \
    InvisibleComment, Empty
from mcmd.script.model.templates import Template
from mcmd.script.model.types_ import InputType
from mcmd.script.parser.errors import ScriptSyntaxError

# subparsers
_double_quoted_string = regex(r'"[^"\\]*(?:\\[\S\s][^"\\]*)*"').map(lambda s: s[1:-1].replace(r'\"', '"'))
_single_quoted_string = regex(r"'[^'\\]*(?:\\[\S\s][^'\\]*)*'").map(lambda s: s[1:-1].replace(r"\'", "'"))
_text = _single_quoted_string.map(Template) | _double_quoted_string.map(Template)
_boolean = string('true').result(True) | string('false').result(False)
_padding = whitespace.optional()
_equals = _padding >> string('=') >> _padding
_bool_assignment = _equals >> _boolean.desc('boolean')
_text_assignment = _equals >> _text.desc('quoted text')
_assignment = _bool_assignment | _text_assignment
_message = _padding >> string(':') >> _padding >> _text
_value_name = regex(r'[a-zA-Z0-9_]+').desc('value name consisting of letters, numbers and/or underscores')
_input_type = from_enum(InputType)
_wait_message = _message | regex('.*').map(lambda s: s.strip()).map(Template)


# $value name [= "text"]
@generate
def _value_declaration():
    yield string('value')
    yield whitespace
    name = yield _value_name
    value = yield _assignment.optional()
    return Value.from_untyped_value(name=name, value=value)


# $input type name [: "message"]
@generate
def _input_declaration():
    yield string('input')
    yield whitespace
    type_ = yield _input_type
    yield whitespace
    name = yield _value_name

    message = yield _message.optional()

    return Input(name=name,
                 type=type_,
                 message=message)


@generate
def _command():
    command_str = yield regex(r'^[^#/\$].*')
    return Command(command=Template(command_str))


# comments (# and //)
@generate
def _visible_comment():
    yield string('#')
    yield _padding
    text = yield regex('.*')
    return VisibleComment(text=Template(text))


_invisible_comment = seq(__start=string('//'),
                         __pad=_padding,
                         comment=regex('.*')).combine_dict(InvisibleComment)

# $wait [: "message"]
# $wait [message]    (backward compatibility)
_wait_declaration = seq(__keyword=string("wait"),
                        message=_wait_message).combine_dict(Wait)

# define declaration parser
_declaration = string('$') >> (_value_declaration | _input_declaration | _wait_declaration)

# define comment parser
_comment = _visible_comment | _invisible_comment

# empty line or line with only whitespace
_empty_line = regex(r'^\s*$').result(Empty())

# top-level parser - the command parser is a fallback parser and will match "everything else"
_statement = _declaration | _comment | _empty_line | _command


def _parse_line(line: str) -> Statement:
    return _statement.parse(line.strip())


def parse(line: Line) -> Statement:
    try:
        return _parse_line(line.string)
    except ParseError as e:
        column = int(str(e).rsplit(':', 1)[-1])
        raise ScriptSyntaxError(str(e), line=line, column=column)
    except TemplateSyntaxError as e:
        raise ScriptSyntaxError(str(e), line=line)



"""This module uses internals of the argparse library to realise functionality that is not (yet) supported."""

import argparse
import textwrap

from mcmd.commands import get_command_names
from mcmd.core.command import CommandType


# noinspection PyProtectedMember
class GroupedHelpFormatter(argparse.RawTextHelpFormatter):
    """Grouping commands is not supported in argparse. This custom help formatter inserts
    headers between the groups."""

    def __init__(self, *args, **kwargs):
        super(GroupedHelpFormatter, self).__init__(*args, **kwargs)
        self._standard_commands = get_command_names(CommandType.STANDARD)
        self._local_commands = get_command_names(CommandType.LOCAL)
        self._meta_commands = get_command_names(CommandType.META)

    def _get_group_header(self, name):
        """Check if this command is the first in one of the groups. If so, return the header for that group."""
        if self._standard_commands[0] == name:
            return textwrap.dedent('''\
                            MOLGENIS commands:
                            ''')
        elif len(self._local_commands) > 0 and self._local_commands[0] == name:
            return textwrap.dedent('''
                            Local commands:
                            ''')
        elif self._meta_commands[0] == name:
            return textwrap.dedent('''
                            meta commands:
                            ''')
        else:
            return False

    def _format_action(self, action):
        """Overrides RawTextHelpFormatter._format_action"""
        help_string = super(GroupedHelpFormatter, self)._format_action(action)

        header = self._get_group_header(action.dest)
        if header:
            help_string = header + help_string
        return help_string


# noinspection PyProtectedMember
def list_subcommands_in_help(parser):
    """Adds a list of subcommands to a command's help string."""
    main_parser = [
        action for action in parser._actions
        if isinstance(action, argparse._SubParsersAction)][0]

    for command in main_parser._choices_actions:
        if command.dest in main_parser.choices:
            subcommands = [
                action for action in main_parser.choices[command.dest]._actions
                if isinstance(action, argparse._SubParsersAction)]
            if len(subcommands) == 0:
                continue
            else:
                subcommands = subcommands[0]

            subs = []
            # noinspection PyUnresolvedReferences
            for subcommand in subcommands._choices_actions:
                subs.append(subcommand.dest)

            command.help = '{}\r- subcommands: {}'.format(command.help,
                                                          ', '.join(subs))



from mcmd.config import config


class McmdError(Exception):
    def __init__(self, message, info: str = None):
        """
        :param message: the error message to show
        :param info: an optional info message that will be shown after the error
        """
        self.message = message
        self.info = info

    def __str__(self):
        return repr(self.message)


class ScriptError(McmdError):
    def __init__(self, message, line: int, info: str = None):
        super().__init__(message, info)
        self.line = line

    @classmethod
    def from_error(cls, error: McmdError, line: int):
        return cls(error.message, line, error.info)


class ConfigError(McmdError):
    def __init__(self, message):
        super().__init__(message="There's an error in the configuration file: {}".format(message))


class MolgenisOfflineError(McmdError):
    def __init__(self):
        super().__init__(message="Can't connect to {}".format(config.url()))



"""Client for the Permission Manager plugin of MOLGENIS (all versions)."""

from urllib.parse import urljoin

from mcmd.core.errors import McmdError
from mcmd.molgenis import api
from mcmd.molgenis.client import post_form, post
from mcmd.molgenis.principals import PrincipalType, to_role_name
from mcmd.molgenis.resources import ResourceType
from mcmd.molgenis.security.permission import Permission


def grant_permission(principal_type: PrincipalType,
                     principal_name: str,
                     resource_type: ResourceType,
                     entity_type_id: str,
                     permission: Permission):
    """Grants permission on one (non-row) resource to one principal."""

    data = {'radio-' + entity_type_id: permission.value}

    if principal_type == PrincipalType.USER:
        data['username'] = principal_name
    elif principal_type == PrincipalType.ROLE:
        principal_name = to_role_name(principal_name)
        data['rolename'] = principal_name
    else:
        raise McmdError('Unknown principal type: %s' % principal_type)

    url = urljoin(api.permission_manager_permissions(),
                  '{}/{}'.format(resource_type.get_resource_name(), principal_type.value))
    post_form(url, data)


def enable_row_level_security(entity_type_id: str):
    """Enables row level security on an entity type."""

    post(api.permission_manager_rls(), data={'id': entity_type_id,
                                             'rlsEnabled': True})


def disable_row_level_security(entity_type_id: str):
    """Disables row level security on an entity type."""

    post(api.permission_manager_rls(), data={'id': entity_type_id,
                                             'rlsEnabled': False})



import shlex
import sys
from typing import List

import attr

from mcmd.args import parser as arg_parser
from mcmd.args.errors import ArgumentSyntaxError
from mcmd.core.errors import McmdError, ScriptError
from mcmd.io import io, ask
from mcmd.io.io import bold, dim
from mcmd.io.logging import get_logger
from mcmd.script.model.lines import ParsedLine
from mcmd.script.model.script import Script
from mcmd.script.model.statements import Value, Input, Wait, VisibleComment, Command, InvisibleComment, \
    Empty
from mcmd.script.model.types_ import InputType, ValueType
from mcmd.script.options import ScriptOptions

log = get_logger()


@attr.s(auto_attribs=True)
class _ScriptExecutionState:
    script: Script
    options: ScriptOptions
    values: dict = attr.Factory(dict)


def run(script: Script, options: ScriptOptions):
    """
    Runs a Script by executing it line by line. Keeps track of the script's state in a _ScriptExecutionState.
    """
    state = _ScriptExecutionState(script=script, options=options, values=options.arguments)

    lines = script.get_lines_with_dependencies(from_line_number=options.start_at)
    _validate_required_args(lines, options.arguments)

    _process_lines(lines, state)


def _validate_required_args(lines: List[ParsedLine], args: dict):
    missing_args = list()
    for line in lines:
        statement = line.statement
        if isinstance(statement, Value) and statement.value is None and statement.name not in args:
            missing_args.append(statement.name)

    if len(missing_args) > 0:
        raise McmdError('Missing required argument(s): {}'.format(', '.join(missing_args)))


def _process_lines(lines: List[ParsedLine], state: _ScriptExecutionState):
    for line in lines:
        try:
            _process_line(line, state)
        except (McmdError, ArgumentSyntaxError) as error:
            _handle_error(error, line_number=line.number, state=state)


def _process_line(line: ParsedLine, state: _ScriptExecutionState):
    statement = line.statement

    if isinstance(statement, Value):
        __set_value(statement, state)
    elif isinstance(statement, Input):
        __ask_input(statement, state)
    elif isinstance(statement, Wait):
        _wait(statement, state)
    elif isinstance(statement, VisibleComment):
        _log_comment(statement, state)
    elif isinstance(statement, InvisibleComment):
        pass
    elif isinstance(statement, Empty):
        pass
    elif isinstance(statement, Command):
        _run_command(statement, state)


def _log_comment(comment: VisibleComment, state: _ScriptExecutionState):
    if state.options.log_comments:
        if len(comment.text.string) == 0:
            io.newline()
        else:
            log.info(comment.text.render(state.values))


def __set_value(assignment: Value, state: _ScriptExecutionState):
    if assignment.value is None or assignment.name in state.values:
        return

    if assignment.type == ValueType.TEMPLATE:
        state.values[assignment.name] = assignment.value.render(state.values)
    else:
        state.values[assignment.name] = assignment.value


def __ask_input(input_: Input, state: _ScriptExecutionState):
    if input_.name in state.values:
        return

    message = input_.message.render(state.values) if input_.message else input_.name

    if input_.type == InputType.TEXT:
        value = ask.input_(message, required=True)
    elif input_.type == InputType.BOOL:
        value = ask.confirm(message)
    elif input_.type == InputType.PASS:
        value = ask.password(message, required=True)
    else:
        raise ValueError('Invalid InputType: {}'.format(input_.type))

    state.values[input_.name] = value


def _wait(wait: Wait, state: _ScriptExecutionState):
    text = '{}: {} {}'.format(bold('Waiting for user'),
                              wait.message.render(state.values),
                              dim('(Press enter to continue)'))
    io.start(text)
    io.wait_for_enter()
    io.succeed()


def _handle_error(error: McmdError, line_number: int, state: _ScriptExecutionState):
    if state.options.exit_on_error and not state.options.dry_run:
        raise ScriptError.from_error(error, line_number)
    else:
        sys.stderr.write('Error on line {}: '.format(str(line_number)))
        io.error(error.message)
        if error.info:
            io.info(error.info)


def _run_command(command: Command, state: _ScriptExecutionState):
    cmd = command.command.render(state.values)
    try:
        sub_args = arg_parser.parse_args(shlex.split(cmd))
    except ArgumentSyntaxError as e:
        raise McmdError(message=str(e))

    _block_nested_scripts(sub_args.command)

    if state.options.dry_run:
        log.info(cmd)
    else:
        setattr(sub_args, 'arg_string', cmd)

        sub_args.func(sub_args, nested=True)


def _block_nested_scripts(command: str):
    if command == 'run':
        raise McmdError('Nesting scripts is not supported', info='If you think it should be, feel free to submit a '
                                                                 'feature request at '
                                                                 'https://github.com/molgenis/molgenis-tools'
                                                                 '-commander/issues!')



from pathlib import Path
from typing import List

from mcmd.config import config
from mcmd.core.context.base_context import Context


class HomeContext(Context):
    """Default context: stores locations to files/folders in the user's home directory: ~/.mcmd"""

    _MCMD_FOLDER = '.mcmd'
    _SCRIPT_FOLDER = 'scripts'
    _BACKUP_FOLDER = 'backups'
    _ISSUE_FOLDER = 'issues'
    _PROPERTIES_FILE = 'mcmd.yaml'
    _HISTORY_FILE = 'history.log'
    _STORAGE_FILE = '.storage.pkl'

    def _get_mcmd_home(self) -> Path:
        mcmd_home = Path().home().joinpath(self._MCMD_FOLDER)
        return self._mkdir_if_not_exists(mcmd_home)

    def get_scripts_folder(self) -> Path:
        scripts_folder = self._get_mcmd_home().joinpath(self._SCRIPT_FOLDER)
        return self._mkdir_if_not_exists(scripts_folder)

    def get_backups_folder(self) -> Path:
        backups_folder = self._get_mcmd_home().joinpath(self._BACKUP_FOLDER)
        return self._mkdir_if_not_exists(backups_folder)

    def get_issues_folder(self) -> Path:
        issues_folder = self._get_mcmd_home().joinpath(self._ISSUE_FOLDER)
        return self._mkdir_if_not_exists(issues_folder)

    def get_history_file(self) -> Path:
        return self._get_mcmd_home().joinpath(self._HISTORY_FILE)

    def get_dataset_folders(self) -> List[Path]:
        return [Path(folder) for folder in config.get('resources', 'dataset_folders')]

    def get_resource_folders(self) -> List[Path]:
        return [Path(folder) for folder in config.get('resources', 'resource_folders')]

    def get_git_folders(self) -> List[Path]:
        return config.git_paths()

    def get_properties_file(self) -> Path:
        return self._get_mcmd_home().joinpath(self._PROPERTIES_FILE)

    def get_storage_file(self) -> Path:
        return self._get_mcmd_home().joinpath(self._STORAGE_FILE)

    @staticmethod
    def _mkdir_if_not_exists(path) -> Path:
        path.mkdir(exist_ok=True)
        return path



import attr

from mcmd.script.model.statements import Statement


@attr.s(frozen=True, auto_attribs=True)
class Line:
    number: int
    string: str


@attr.s(frozen=True, auto_attribs=True)
class ParsedLine:
    """
    A parsed line contains the line number, original line string and the parsed Statement. A parsed line may have
    dependencies on other lines (through Templates). These are exposed through the dependencies property.
    """

    raw: str
    number: int
    statement: Statement



"""
Alter values
"""
import json
import textwrap
from argparse import RawDescriptionHelpFormatter
from pathlib import Path

from mcmd.commands._registry import arguments
from mcmd.core.command import command
from mcmd.core.context import context
from mcmd.core.errors import McmdError
from mcmd.io import io
from mcmd.io.io import highlight
from mcmd.molgenis import api
from mcmd.molgenis.client import get, put
from mcmd.utils import files


# =========
# Arguments
# =========


@arguments('set')
def add_arguments(subparsers):
    p_set = subparsers.add_parser('set',
                                  help='update a single entity value',
                                  formatter_class=RawDescriptionHelpFormatter,
                                  description=textwrap.dedent(
                                      """
                                      Updates a single value in any entity. When updating a settings entity it is not 
                                      needed to supply the row's identifier (with '--for').

                                      example usage:
                                        # Updating a value in a row:
                                        mcmd set my_patients name john --for patient_1
                                        
                                        # Updating a setting:
                                        mcmd set app title "My MOLGENIS"
                                        
                                        # Using the contents of a file as the value
                                        mcmd set app molgenis_menu --from-resource menu.json
                                        mcmd set app footer --from-path /my/path/footer.txt
                                      """
                                  ))

    p_set.add_argument('type',
                       type=str,
                       help="either 1) a simple name of a settings entity (app, mail, opencpu, auth or dataexplorer), "
                            "or 2) the ID of any entity type")

    p_set.add_argument('attribute',
                       type=str,
                       help="the attribute to set")

    p_set.add_argument('value',
                       type=str,
                       help="the value to set the attribute to")

    p_set_file = p_set.add_mutually_exclusive_group()
    p_set_file.add_argument('--from-resource', '-r',
                            action='store_true',
                            help='flag to specify that the value should come from the contents of the specified '
                                 'resource')
    p_set_file.add_argument('--from-path', '-p',
                            action='store_true',
                            help='flag to specify that the value should come from the contents of the specified '
                                 'absolute file path')

    p_set.add_argument('--for', '-i',
                       metavar='ENTITY ID',
                       type=str,
                       dest='for_',
                       help="The id of the row you want to alter")

    p_set.set_defaults(func=set_,
                       write_to_history=True)


# =======
# Globals
# =======


_SETTING_SYNONYMS = {'mail': 'sys_set_MailSettings',
                     'opencpu': 'sys_set_OpenCpuSettings',
                     'app': 'sys_set_app',
                     'auth': 'sys_set_auth',
                     'dataexplorer': 'sys_set_dataexplorer'
                     }


# =======
# Methods
# =======

@command
def set_(args):
    """
    set sets the specified row of the specified table (or setting of specified settings table) to the specified value
    :param args: command line arguments containing: the settings type, the setting to set, and the value to set it to,
    if not a setting also the --for (which row to alter)
    if --from-path or --from-resource is passed the value is assumed to be a file containing the value data to be set
    :return: None
    """
    value = args.value
    value_desc = highlight(value)

    if args.from_path:
        path = Path(args.value)
        value = files.read_file(path)
        value_desc = 'contents of {}'.format(highlight(args.value))
    elif args.from_resource:
        path = files.select_file_from_folders(context().get_resource_folders(), args.value)
        value = files.read_file(path)
        value_desc = 'contents of {}'.format(highlight(args.value))

    if args.for_:
        entity = args.type
        row = args.for_
        io.start(
            'Updating {} of {} for id {} to {}'.format(highlight(args.attribute), highlight(args.type),
                                                       highlight(args.for_), value_desc))
    else:
        entity = _get_settings_entity(args.type)
        io.start(
            'Updating {} of {} settings to {}'.format(highlight(args.attribute), highlight(args.type),
                                                      value_desc))
        row = _get_first_row_id(entity)

    url = api.rest1('{}/{}/{}'.format(entity, row, args.attribute))
    put(url, json.dumps(value))


def _get_settings():
    molgenis_settings = get(api.rest2('sys_md_EntityType'),
                            params={
                                'q': 'extends==sys_set_settings',
                                'attrs': '~id'
                            }).json()['items']
    return [setting['id'] for setting in molgenis_settings]


def _get_settings_entity(setting):
    if setting.lower() in _SETTING_SYNONYMS:
        return _SETTING_SYNONYMS[setting.lower()]
    else:
        possible_settings = _get_settings()
        settings_entity = [selected_setting for selected_setting in possible_settings if
                           setting.lower() == selected_setting.lower()]
        if len(settings_entity) == 1:
            return settings_entity[0]
        else:
            raise McmdError('Setting [{}] is not a valid settings entity. If you meant to alter'.format(setting) +
                            ' another table, please provide the --for option to specify the row to alter.')


def _get_first_row_id(entity):
    settings = get(api.rest2(entity),
                   params={
                       'attrs': '~id'
                   }).json()['items']
    return settings[0]['id']



from os import path
from pathlib import Path
from typing import List

import mcmd.io.ask
from mcmd.core.errors import McmdError
from mcmd.io import io


def get_file_name_from_path(file_path: str) -> str:
    """
    get_file_name returns a file name from a path to the file
    :param file_path: path to the file (i.e. /Users/henk/Desktop/example.xlsx)
    :return: filename (i.e. example.xlsx)
    """
    return path.basename(file_path)


def get_files(folders: List[Path]) -> List[Path]:
    """
    loops through specified folders and appends their files to a list
    :param folders: a list of paths to folders
    :return: files: a dictionary with as key a file and as value a list of the paths that lead to the file
    """
    files = list()
    for folder in folders:
        if not folder.is_dir():
            io.warn("Folder %s doesn't exist" % folder)

        for file in list(folder.glob('*.*')):
            files.append(file)
    return files


def read_file(file: Path) -> str:
    """
    read_file reads the file data into a string
    :param file: file to read from
    :return: file contents
    """
    try:
        with file.open() as file_handle:
            content = file_handle.read()
    except OSError as e:
        raise McmdError('Error reading file: {}'.format(str(e)))
    return content


def read_file_lines(file: Path) -> List[str]:
    """
    read_file reads the file data into a list of strings
    :param file: file to read from
    :return: list of lines
    """
    try:
        with file.open() as file_handle:
            content = file_handle.read().splitlines()
    except OSError as e:
        raise McmdError('Error reading file: {}'.format(str(e)))
    return content


def select_file_from_folders(folders: List[Path], file_name: str) -> Path:
    """
    Selects a file from a list of folders. File name can be supplied with or without extension. When there are multiple
    matches, the user will be asked to choose one.
    :param folders: a list of folders
    :param file_name: the name of the file to get the path of, with or without extension
    :return: the selected path

    :exception McmdError if selected file was not found
    """
    files = get_files(folders)
    return select_file(files, file_name)


def select_file(files: List[Path], file_name: str) -> Path:
    """
    select_path selects the path from a list of paths, asks user's input when result is ambiguous
    :param files: a list of files
    :param file_name: the name of the file to get the path of, with or without extension
    :return: the selected path

    :exception McmdError if selected file was not found
    """

    matches = list()
    for file in files:
        if file_name == file.stem or file_name == file.name:
            matches.append(file)

    if len(matches) > 0:
        if len(matches) > 1:
            file_path = _choose_file(matches, file_name)
        else:
            file_path = matches[0]
    else:
        raise McmdError('No file found for {}'.format(file_name))
    return file_path


def _choose_file(paths: List[Path], name: str):
    """
    _choose_file asks the user to choose a file out of the possible file paths
    :param paths: the list of paths to the file
    :param name: the filename
    :return: the selected path
    """
    choices = [str(file_path) for file_path in paths]
    answer = mcmd.io.ask.multi_choice('Multiple files found for %s. Pick one:' % name, choices)
    return Path(answer)



import argparse
from io import StringIO

from mcmd.args.errors import ArgumentSyntaxError


class RaisingArgumentParser(argparse.ArgumentParser):
    """
    When an error occurs, the default ArgumentParser prints the usage + error message directly to stderr and then exits
    the program. This subclass raises an ArgumentSyntaxError instead so that the caller can decide what to do when
    parsing fails.
    """

    def error(self, message):
        message = self.prog + ': ' + message
        raise ArgumentSyntaxError(message=message, usage=self.get_usage_as_string())

    def get_usage_as_string(self):
        with StringIO() as usage_io:
            self.print_usage(usage_io)
            return usage_io.getvalue()



"""
Houses the @arguments decorator, which registers each command's arguments.
"""
from collections import defaultdict

from mcmd.core.command import CommandType

_ARGUMENT_ADDER_REGISTRY = defaultdict(dict)


def get_argument_adders(type_: CommandType):
    return [func for name, func in sorted(_ARGUMENT_ADDER_REGISTRY[type_].items())]


def get_command_names(type_: CommandType):
    return [name for name, func in sorted(_ARGUMENT_ADDER_REGISTRY[type_].items())]


def arguments(command_name, type_=CommandType.STANDARD):
    """Command argument adder registration decorator."""

    def decorator(func):
        _ARGUMENT_ADDER_REGISTRY[type_][command_name] = func
        return func

    return decorator



from os import path as os_path
from pathlib import Path
from urllib.parse import urljoin

import polling
import requests

import mcmd.config.config as config
import mcmd.io.ask
from mcmd.commands._registry import arguments
from mcmd.core.command import command
from mcmd.core.context import context
from mcmd.core.errors import McmdError
from mcmd.github import client as github
from mcmd.io import io
from mcmd.io.io import highlight
from mcmd.molgenis import api
from mcmd.molgenis.client import post_file, get, post
from mcmd.utils import files

# =========
# Arguments
# =========

# Store a reference to the parser so that we can show an error message for the custom validation rule
_p_import = None


@arguments('import')
def add_arguments(subparsers):
    global _p_import
    _p_import = subparsers.add_parser('import',
                                      help='import a dataset')
    _p_import.set_defaults(func=import_,
                           write_to_history=True)
    _p_import.add_argument('resource',
                           nargs='?',
                           help='the resource to import - depending on the other options this can be a path, '
                                'file name, or URL')
    p_import_source = _p_import.add_mutually_exclusive_group()
    p_import_source.add_argument('--from-path', '-p',
                                 action='store_true',
                                 help='import a file the old school way: by path')
    p_import_source.add_argument('--from-issue', '-i',
                                 metavar='NUMBER',
                                 help="import a file attachment from a GitHub issue - optionally supply the file name "
                                      "of the attachment")
    p_import_source.add_argument('--from-url', '-u',
                                 action='store_true',
                                 help='import a file from a URL - uses the importByUrl endpoint of the MOLGENIS '
                                      'importer, without downloading the file first')
    _p_import.add_argument('--in',
                           dest='to_package',
                           type=str,
                           metavar='PACKAGE_ID',
                           help='the package to import to')
    _p_import.add_argument('--as',
                           dest='entity_type_id',
                           type=str,
                           metavar='ENTITY_TYPE_ID',
                           help='the id of the entity type (only used when importing VCF files)')
    _p_import.add_argument('--with-action', '-a',
                           dest='import_action',
                           type=str,
                           choices=['add', 'add_update_existing', 'update'],
                           help='strategy to use when importing')
    return _p_import


# =========
# Globals
# =========

_IMPORT_ACTIONS = {'.owl': 'add',
                   '.obo': 'add',
                   '.vcf': 'add'}


# =======
# Methods
# =======

@command
def import_(args):
    _validate_args(args)
    _choose_import_method(args)


def _validate_args(args):
    """The 'resource' argument can't be made required at the parser level because the '--from-issue' argument can be
    used with or without specifying a file name."""
    if not args.resource and not args.from_issue:
        _p_import.error("the following argument is required: resource")


def _choose_import_method(args):
    if args.from_path:
        _import_from_path(args)
    elif args.from_url:
        _import_from_url(args)
    elif args.from_issue:
        _import_from_issue(args)
    else:
        _import_from_quick_folders(args)


def _import_from_url(args):
    file_url = args.resource
    file_name = file_url.split("/")[-1]
    io.start('Importing from URL %s' % highlight(file_url))

    if args.import_action:
        action = args.import_action
    else:
        action = _get_import_action(file_name)

    params = {'action': action,
              'metadataAction': 'upsert'}

    if args.to_package:
        params['packageId'] = args.to_package

    params['url'] = file_url

    response = post(api.import_by_url(), params=params)
    import_run_url = urljoin(config.get('host', 'selected'), response.text)
    status, message = _poll_for_completion(import_run_url)
    if status == 'FAILED':
        raise McmdError(message)


def _import_from_quick_folders(args):
    file_name = os_path.splitext(args.resource)[0]
    path = files.select_file_from_folders(folders=context().get_git_folders() + context().get_dataset_folders(),
                                          file_name=file_name)
    _do_import(path, args.to_package, args.entity_type_id, args.import_action)


def _import_from_issue(args):
    issue_num = args.from_issue
    attachment = _select_attachment(issue_num, args.resource)
    file_path = _download_attachment(attachment, issue_num)
    _do_import(file_path, args.to_package, args.entity_type_id, args.import_action)


def _import_from_path(args):
    file = Path(args.resource)
    if not file.is_file():
        raise McmdError("File %s doesn't exist" % str(file.resolve()))
    _do_import(file, args.to_package, args.entity_type_id, args.import_action)


def _select_attachment(issue_num, wanted_attachment):
    """Gets attachments from a GitHub issue. If wanted_attachment is specified it will try to select that attachment."""
    attachments = github.get_attachments(issue_num)
    if len(attachments) == 0:
        raise McmdError("Issue #%s doesn't contain any files" % issue_num)

    if wanted_attachment:
        selected = [a for a in attachments if a.name == wanted_attachment]
        if len(selected) == 0:
            raise McmdError('There are no attachments named %s.' % wanted_attachment)
        if len(selected) > 1:
            raise McmdError('Multiple attachments with name %s found.' % wanted_attachment)
        return selected[0]
    else:
        if len(attachments) > 1:
            return _choose_attachment(attachments)
        else:
            return attachments[0]


def _create_attachment_map(attachments):
    """"Creates a dict of name/attachment pairs. If the name isn't unique, uses the identifier as key. Example:
            1234/example.xls
            4567/example.xls
            other_example.xls
    """
    names = [a.name for a in attachments]

    attachment_map = dict()
    for a in attachments:
        if names.count(a.name) > 1:
            attachment_map[a.id] = a
        else:
            attachment_map[a.name] = a

    return attachment_map


def _choose_attachment(attachments):
    """Let user choose from multiple attachments. Attachments with duplicate names will display their id."""
    attachment_map = _create_attachment_map(attachments)
    choices = list(attachment_map.keys())

    answer = mcmd.io.ask.multi_choice('Multiple attachments found. Choose which one to import:', choices)
    return attachment_map[answer]


def _download_attachment(attachment, issue_num):
    issue_folder = context().get_issues_folder().joinpath(issue_num)
    issue_folder.mkdir(parents=True, exist_ok=True)
    file_path = issue_folder.joinpath(attachment.name)

    if file_path.exists():
        overwrite = mcmd.io.ask.confirm('File %s already exists. Re-download?' % file_path.name)
        if not overwrite:
            return file_path

    io.start('Downloading %s from GitHub issue %s' % (highlight(attachment.name), highlight('#' + issue_num)))
    try:
        r = requests.get(attachment.url)
        r.raise_for_status()
        with file_path.open('wb') as f:
            f.write(r.content)
    except (OSError, requests.RequestException, requests.HTTPError) as e:
        raise McmdError('Error downloading GitHub attachment: %s' % str(e))
    io.succeed()
    return file_path


def _do_import(file_path, package, entity_type_id, import_action):
    io.start('Importing %s' % (highlight(file_path.name)))

    if import_action:
        action = import_action
    else:
        action = _get_import_action(file_path.name)

    params = {'action': action,
              'metadataAction': 'upsert'}

    if package:
        params['packageId'] = package
    if entity_type_id:
        params['entityTypeId'] = entity_type_id

    response = post_file(api.import_(), file_path.resolve(), params)
    import_run_url = urljoin(config.get('host', 'selected'), response.text)
    status, message = _poll_for_completion(import_run_url)
    if status == 'FAILED':
        raise McmdError(message)


def _get_import_action(file_name):
    file_name = file_name.rstrip('.gz')
    file_name = file_name.rstrip('.zip')
    for file_type, action in _IMPORT_ACTIONS.items():
        if file_name.endswith(file_type):
            return action
    return config.get('settings', 'import_action')


def _poll_for_completion(url):
    def step_function(step):
        """Increases time between polls with one second each time, with a maximum of 10 seconds."""
        step += 1
        return min(step, 10)

    polling.poll(lambda: get(url).json()['status'] != 'RUNNING',
                 step=0,
                 step_function=step_function,
                 poll_forever=True)
    import_run = get(url).json()
    return import_run['status'], import_run['message']



import sys

from mcmd.args import _factory

_parser = None


def _get_parser():
    global _parser
    if not _parser:
        _parser = _factory.create_parser()

    return _parser


def parse_args(arg_list):
    args = _get_parser().parse_args(arg_list)
    _show_help(args, arg_list)
    return args


def _show_help(args, arg_list):
    if not args.command:
        _print_help()
        exit(1)
    elif _is_intermediate_subcommand(args):
        # we can't access the subparser from here, so we parse the arguments again with the --help flag
        arg_list.append('--help')
        parse_args(arg_list)
        exit(1)


def _print_help():
    _get_parser().print_help(sys.stderr)
    exit(1)


def _is_intermediate_subcommand(args):
    """
    Some commands have nested subcommands. These intermediate commands are not executable and don't have a 'func'
    property.
    For example:
    > mcmd add user
    Here, 'add' is the intermediate command.
    """
    return not hasattr(args, 'func')



from typing import List, Set, Dict

import attr

from mcmd.script.model.lines import ParsedLine


@attr.s(frozen=True, auto_attribs=True)
class Script:
    lines: List[ParsedLine]
    _dependencies: Dict[ParsedLine, Set[ParsedLine]]

    def get_lines_with_dependencies(self, from_line_number: int) -> List[ParsedLine]:
        """
        Gets all the lines in the script from a certain line_number and up. If the lines depend on lines before
        line_number, then those lines will be included as well. This is done recursively, so all dependencies are
        included.
        """
        if from_line_number < 2:
            return list(self.lines)
        else:
            lines_subset = {line for line in self.lines if line.number >= from_line_number}

            dependencies = set()
            for line in lines_subset:
                dependencies |= self._get_deep_dependencies(line)

            total_lines = (lines_subset | dependencies)
            return sorted(total_lines, key=lambda l: l.number)

    def _get_deep_dependencies(self, line: ParsedLine) -> Set[ParsedLine]:
        dependencies = set()
        dependencies.add(line)
        for dependency in self._dependencies.get(line, set()):
            dependencies |= self._get_deep_dependencies(dependency)
        return dependencies



from typing import List

from mcmd.script.model.lines import Line


def _ends_with_continuation_symbol(string: str) -> bool:
    return string.endswith(" \\") or string.endswith("\t\\")


def _strip_continuation_symbol(string: str) -> str:
    return string[:-1].strip()


def combine_multi_lines(lines: List[str]) -> List[Line]:
    """
    Combines multi-lines (lines that end with '\') into single Line objects.
    """

    combined_lines = list()

    if _ends_with_continuation_symbol(lines[-1]):
        lines[-1] = _strip_continuation_symbol(lines[-1])

    # iterate the lines backwards
    for i in range(len(lines) - 1, -1, -1):
        if i > 0:
            previous_line = lines[i - 1]
            continued = _ends_with_continuation_symbol(previous_line)
        else:
            continued = False

        if continued:
            # strip the continuation symbol of the previous line and add this line to it
            lines[i - 1] = _strip_continuation_symbol(lines[i - 1]) + ' ' + lines[i]
        else:
            # commit this line
            combined_lines.append(Line(number=i + 1, string=lines[i]))

    combined_lines.reverse()
    return combined_lines



import mimetypes
import textwrap
from argparse import RawDescriptionHelpFormatter
from os import path as os_path
from typing import List

from mcmd.commands._registry import arguments
from mcmd.core.command import command
from mcmd.core.compatibility import version, deprecated
from mcmd.core.context import context
from mcmd.core.errors import McmdError
from mcmd.io import io
from mcmd.io.io import highlight
from mcmd.molgenis import api
from mcmd.molgenis.client import post, get, post_files
from mcmd.molgenis.principals import to_role_name
from mcmd.utils import files as file_utils

# =========
# Arguments
# =========

# Store a reference to the parser so that we can show an error message for the custom validation rule
p_add_theme = None


@arguments('add')
def add_arguments(subparsers):
    global p_add_theme
    p_add = subparsers.add_parser('add',
                                  help='add and upload resources',
                                  description="Run 'mcmd add group -h' or 'mcmd add user -h' to view the help for those"
                                              " subcommands")
    p_add_subparsers = p_add.add_subparsers(dest="type", metavar='')

    p_add_group = p_add_subparsers.add_parser('group',
                                              help='add a group')
    p_add_group.set_defaults(func=add_group,
                             write_to_history=True)
    p_add_group.add_argument('name',
                             type=str,
                             help="the group's name")

    p_add_user = p_add_subparsers.add_parser('user',
                                             help='add a user')
    p_add_user.set_defaults(func=add_user,
                            write_to_history=True)
    p_add_user.add_argument('username',
                            type=str,
                            help="the user's name")
    p_add_user.add_argument('--set-password', '-p',
                            metavar='PASSWORD',
                            type=str,
                            help="the user's password")
    p_add_user.add_argument('--with-email', '-e',
                            metavar='EMAIL',
                            type=str,
                            help="the user's e-mail address")
    p_add_user.add_argument('--is-inactive', '-a',
                            action='store_true',
                            help="make user inactive")
    p_add_user.add_argument('--is-superuser', '-s',
                            action='store_true',
                            help="make user superuser")
    p_add_user.add_argument('--change-password', '-c',
                            action='store_true',
                            help="set change password to true for user")

    p_add_role = p_add_subparsers.add_parser('role',
                                             help='add a role',
                                             formatter_class=RawDescriptionHelpFormatter,
                                             description=textwrap.dedent(
                                                 """
                                                 Add a (group) role. 
    
                                                 Note: since MOLGENIS 8.3 role names are case sensitive and need to 
                                                 be typed exactly as-is. (Before 8.3 all role names will be upper 
                                                 cased automatically). 
    
                                                 Example usage:
                                                   mcmd add role COUNTER
                                                   mcmd add role gcc_COUNTER --to-group gcc --includes COUNTER                                      
                                                 """
                                             ))
    p_add_role.set_defaults(func=add_role,
                            write_to_history=True)
    p_add_role.add_argument('rolename',
                            help="the role's name")
    p_add_role.add_argument('--to-group', '-g',
                            dest='group',
                            metavar='GROUP NAME',
                            help='the group to add the role to')
    p_add_role.add_argument('--includes', '-i',
                            metavar='ROLE NAMES',
                            nargs='+',
                            help='the role(s) that this role includes (if more than one, separated by a whitespace)')

    p_add_package = p_add_subparsers.add_parser('package',
                                                help='add a package')
    p_add_package.set_defaults(func=add_package,
                               write_to_history=True)
    p_add_package.add_argument('id',
                               type=str,
                               help="the id of the Package")
    p_add_package.add_argument('--in',
                               type=str,
                               dest='parent',
                               help="the id of the parent")

    p_add_token = p_add_subparsers.add_parser('token',
                                              help='add a token')
    p_add_token.set_defaults(func=add_token,
                             write_to_history=True)
    p_add_token.add_argument('user',
                             type=str,
                             help="the user to give the token to")
    p_add_token.add_argument('token',
                             type=str,
                             help="the token")

    p_add_theme = p_add_subparsers.add_parser('theme',
                                              help='upload a bootstrap theme (deprecated since 8.6)',
                                              formatter_class=RawDescriptionHelpFormatter,
                                              description=textwrap.dedent(
                                                  """
                                                  Add a CSS theme. 

                                                  Deprecated since 8.6. The CSS file(s) should be hosted somewhere else.

                                                  Example usage:
                                                    mcmd add theme --bootstrap3 theme3.css --bootstrap4 theme4.css
                                                  """
                                              )
                                              )
    p_add_theme.set_defaults(func=add_theme,
                             write_to_history=True)
    p_add_theme.add_argument('--from-path', '-p',
                             action='store_true',
                             help='add a bootstrap theme by specifying a path')

    required_named = p_add_theme.add_argument_group('required named arguments')
    required_named.add_argument('--bootstrap3', '-3',
                                type=str,
                                metavar='STYLESHEET',
                                help="the bootstrap3 css theme file (when not specified, the default molgenis theme "
                                     "will be applied on bootstrap3 pages)")

    p_add_theme.add_argument('--bootstrap4', '-4',
                             type=str,
                             metavar='STYLESHEET',
                             help="the bootstrap4 css theme file (when not specified, the default molgenis theme will "
                                  "be applied on bootstrap4 pages)")

    p_add_logo = p_add_subparsers.add_parser('logo',
                                             help='upload a logo to be placed on the left top of the menu')
    p_add_logo.set_defaults(func=add_logo,
                            write_to_history=True)
    p_add_logo.add_argument('--from-path', '-p',
                            action='store_true',
                            help='add a logo by specifying a path')
    p_add_logo.add_argument('logo',
                            type=str,
                            help="the image you want to use as logo")


# =======
# Methods
# =======

@command
def add_user(args):
    io.start('Adding user %s' % highlight(args.username))

    password = args.set_password if args.set_password else args.username
    email = args.with_email if args.with_email else args.username + '@molgenis.org'
    active = not args.is_inactive
    superuser = args.is_superuser
    ch_pwd = args.change_password

    post(api.rest1('sys_sec_User'),
         data={'username': args.username,
               'password_': password,
               'changePassword': ch_pwd,
               'Email': email,
               'active': active,
               'superuser': superuser
               })


@command
def add_role(args):
    role_name = to_role_name(args.rolename)
    io.start('Adding role {}'.format(highlight(role_name)))

    role = {'name': role_name,
            'label': role_name}

    if args.includes:
        role_names = [to_role_name(name) for name in args.includes]
        role['includes'] = _get_role_ids(role_names)

    if args.group:
        group_name = _to_group_name(args.group)
        role['group'] = _get_group_id(group_name)

    data = {'entities': [role]}
    post(api.rest2('sys_sec_Role'), data=data)


def _get_group_id(group_name) -> str:
    groups = get(api.rest2('sys_sec_Group'),
                 params={
                     'attrs': 'id',
                     'q': 'name=={}'.format(group_name)
                 }).json()['items']
    if len(groups) == 0:
        raise McmdError('No group found with name {}'.format(groups))
    else:
        return groups[0]['id']


def _get_role_ids(role_names) -> List[str]:
    roles = get(api.rest2('sys_sec_Role'),
                params={
                    'attrs': 'id,name',
                    'q': 'name=in=({})'.format(','.join(role_names))
                }).json()['items']

    name_to_id = {role['name']: role['id'] for role in roles}
    not_found = list()
    for role_name in role_names:
        if role_name not in name_to_id:
            not_found.append(role_name)

    if len(not_found) > 0:
        raise McmdError("Couldn't find role(s) {}".format(' and '.join(not_found)))
    else:
        return list(name_to_id.values())


@command
def add_group(args):
    group_name = _to_group_name(args.name)
    io.start('Adding group %s' % highlight(group_name))
    post(api.group(), data={'name': group_name, 'label': args.name})


@version('7.0.0')
def _to_group_name(group_input: str):
    """Before 8.3.0 all group names are lower case."""
    return group_input.lower()


@version('8.3.0')
def _to_group_name(group_input: str):
    """Since 8.3.0 group names are case sensitive."""
    return group_input


@command
def add_package(args):
    io.start('Adding package %s' % highlight(args.id))

    data = {'id': args.id,
            'label': args.id}

    if args.parent:
        data['parent'] = args.parent

    post(api.rest1('sys_md_Package'), data=data)


@command
def add_token(args):
    io.start('Adding token %s for user %s' % (highlight(args.token), highlight(args.user)))

    user = get(api.rest2('sys_sec_User'),
               params={
                   'attrs': 'id',
                   'q': 'username=={}'.format(args.user)
               })
    if user.json()['total'] == 0:
        raise McmdError('Unknown user %s' % args.user)

    user_id = user.json()['items'][0]['id']

    data = {'User': user_id,
            'token': args.token}

    post(api.rest1('sys_sec_Token'), data=data)


@command
@deprecated(since='8.6.0',
            action='Adding themes',
            info='Themes should be hosted somewhere else. Read '
                 'https://molgenis.gitbook.io/molgenis/configuration/guide-customize#style-theme to learn more.')
def add_theme(args):
    """
    add_theme adds a theme to the stylesheet table
    :param args: commandline arguments containing bootstrap3_theme and optionally bootstrap4_theme
    :return: None
    """
    _validate_args(args)
    valid_types = {'text/css'}
    api_ = api.add_theme()
    bs3_name = args.bootstrap3
    bs4 = args.bootstrap4
    paths = [bs3_name]
    names = ['bootstrap3-style']
    if bs4:
        paths.append(bs4)
        names.append('bootstrap4-style')
        bs4_name = file_utils.get_file_name_from_path(bs4)
        io.start(
            'Adding bootstrap 3 theme {} and bootstrap 4 theme {} to bootstrap themes'.format(
                highlight(bs3_name),
                highlight(bs4_name)))
    else:
        io.start(
            'Adding bootstrap 3 theme {} to bootstrap themes'.format(
                highlight(bs3_name)))
    if not args.from_path:
        paths = [_get_path_from_quick_folders(theme) for theme in paths]
    files = _prepare_files_for_upload(paths, names, valid_types)
    post_files(files, api_)


@command
def add_logo(args):
    """
    add_logo uploads a logo to add to the left top of the menu
    :param args: commandline arguments containing path to logo
    :return: None
    """
    api_ = api.logo()
    valid_types = {'image/jpeg', 'image/png', 'image/gif'}
    logo = [args.logo]
    if not args.from_path:
        io.start('Adding logo from path {}'.format(highlight(args.logo)))
        logo = [_get_path_from_quick_folders(args.logo)]
    else:
        io.start('Adding logo {}'.format(highlight(args.logo)))
    files = _prepare_files_for_upload(logo, ['logo'], valid_types)
    post_files(files, api_)


def _prepare_files_for_upload(paths, names, valid_content_types):
    """
    _prepare_files_for_upload takes the paths to the files to upload, the names of them and a list of valid content
    types to translate to content types and generates a dictionary which can be uploaded in a post request
    :param paths: a list of paths to the files to upload
    :param names: the names of files to upload
    :param valid_content_types: set of the possible valid content types
    :return: a dictionary with as key the name of the file and as value a tuple with: filename, file to upload, and
    content type

    :exception McmdError when the file on the given path does not exist and when the content type of the file is
    invalid.
    """
    files = {}
    for name, path in zip(names, paths):
        file_name = file_utils.get_file_name_from_path(path)
        content_type = mimetypes.guess_type(path)[0]
        if not os_path.exists(path):
            raise McmdError(
                'File [{}] does not exist on path [{}]'.format(file_name, path.strip(file_name)))
        elif content_type in valid_content_types:
            try:
                files[name] = (file_name, open(path, 'rb'), content_type)
            except FileNotFoundError:
                raise McmdError(
                    'File [{}] does not exist on path [{}]'.format(file_name, path.strip(file_name)))
        else:
            raise McmdError(
                'File [{}] does not have valid content type [{}], '
                'content type should be in {}'.format(file_name,
                                                      content_type,
                                                      valid_content_types))
    return files


def _get_path_from_quick_folders(file_name):
    file_name = os_path.splitext(file_name)[0]
    path = file_utils.select_file_from_folders(context().get_resource_folders(), file_name)
    return str(path)


def _validate_args(args):
    """To make the bootstrap themes as backwards compatible as possible make both bootstrap3 and 4 named, but only
    bootstrap3 required, in future 4 will be required and 3 will be removed eventually."""
    if args.type == 'theme' and not args.bootstrap3:
        p_add_theme.error("the following argument is required: bootstrap3")



import mcmd.io.ask
from mcmd.commands._registry import arguments
from mcmd.core.context import context
from mcmd.core import history
from mcmd.core.command import command, CommandType
from mcmd.core.errors import McmdError
from mcmd.io import io
from mcmd.io.ask import confirm
from mcmd.io.io import highlight
from mcmd.io.logging import get_logger


# =========
# Arguments
# =========

@arguments('script', CommandType.META)
def add_arguments(subparsers):
    p_script = subparsers.add_parser('script',
                                     help="do actions involving scripts")
    p_script.set_defaults(func=script,
                          write_to_history=False)
    p_script_action = p_script.add_mutually_exclusive_group()
    p_script_action.add_argument('--create',
                                 action='store_true',
                                 help='create a script from the history (this is the default action)')
    p_script_action.add_argument('--list', '-l',
                                 action='store_true',
                                 help='list the stored scripts')
    p_script_action.add_argument('--delete', '-D',
                                 metavar='SCRIPT NAME',
                                 type=str,
                                 help='remove a script')
    p_script_action.add_argument('--read', '-r',
                                 metavar='SCRIPT NAME',
                                 type=str,
                                 help='read the contents of a script')
    p_script.add_argument('--number', '-n',
                          type=int,
                          default=10,
                          help='number of lines of history to choose from (default: 10)')
    p_script.add_argument('--show-fails', '-f',
                          action='store_true',
                          help='also show the failed commands from history (disabled by default)')


# =======
# Globals
# =======

log = get_logger()


# =======
# Methods
# =======

@command
def script(args):
    if args.list:
        _list_scripts()
    elif args.delete:
        _remove_script(args.delete)
    elif args.read:
        _read_script(args.read)
    else:
        _create_script(args)


def _remove_script(script_name):
    path = context().get_scripts_folder().joinpath(script_name)
    _check_script_exists(path)
    try:
        io.start('Removing script %s' % highlight(script_name))
        path.unlink()
    except OSError as e:
        raise McmdError('Error removing script: %s' % str(e))


def _read_script(script_name):
    path = context().get_scripts_folder().joinpath(script_name)
    _check_script_exists(path)
    try:
        with path.open() as f:
            for line in f.readlines():
                log.info(line.strip())
    except OSError as e:
        raise McmdError('Error reading script: %s' % str(e))


def _list_scripts():
    for path in context().get_scripts_folder().iterdir():
        if not path.name.startswith('.'):
            log.info(path.name)


def _create_script(args):
    lines = history.read(args.number, args.show_fails)
    if len(lines) == 0:
        log.info('History is empty.')
        return

    options = [line[1] for line in lines]
    commands = mcmd.io.ask.checkbox('Pick the lines that will form the script:', options)
    file_name = _input_script_name()
    try:
        with open(str(context().get_scripts_folder().joinpath(file_name)), 'w') as script_file:
            for cmd in commands:
                script_file.write(cmd + '\n')
    except OSError as e:
        raise McmdError("Error writing to script: %s" % str(e))


def _check_script_exists(path):
    if not path.exists():
        raise McmdError("Script %s doesn't exist" % path.name)


def _input_script_name():
    file_name = ''
    while not file_name:
        name = mcmd.io.ask.input_('Supply the name of the script:')
        if context().get_scripts_folder().joinpath(name).exists():
            overwrite = confirm('%s already exists. Overwrite?' % name)
            if overwrite:
                file_name = name
        else:
            file_name = name

    return file_name



import textwrap
from argparse import RawDescriptionHelpFormatter

import mcmd.config.config as config
from mcmd.commands._registry import arguments
from mcmd.core.command import command, CommandType
from mcmd.core.errors import McmdError
from mcmd.io import io, ask
from mcmd.io.io import highlight


# =========
# Arguments
# =========

@arguments('config', CommandType.META)
def add_arguments(config_subparsers):
    config_ = config_subparsers.add_parser('config',
                                           help='change the configuration of MOLGENIS Commander',
                                           formatter_class=RawDescriptionHelpFormatter,
                                           description=textwrap.dedent(
                                               """
                                               Changes values in the configuration file.
    
                                               example usage:
                                                 # Adding a host interactively or with arguments
                                                 mcmd config add host
                                                 mcmd config add host --url http://x --username admin --password admin
    
                                                 # Switching to another host interactively or with arguments
                                                 mcmd config set host
                                                 mcmd config set host http://localhost
    
                                                 # Enabling and disabling non-interactive mode
                                                 mcmd config set non-interactive
                                                 mcmd config set interactive
    
                                                 # Setting the default import action interactively and with arguments
                                                 mcmd config set import-action
                                                 mcmd config set import-action --action add_update_existing
                                               """
                                           ))
    config_subparsers = config_.add_subparsers(dest='type', metavar='')

    set_ = config_subparsers.add_parser('set',
                                        help='set values in the configuration file')

    set_subparsers = set_.add_subparsers(metavar='')
    set_host = set_subparsers.add_parser('host',
                                         help='select a host')
    set_host.set_defaults(func=config_set_host,
                          write_to_history=False)
    set_host.add_argument('url',
                          nargs='?',
                          help='the URL of the host (Optional)')

    set_import_action = set_subparsers.add_parser('import-action',
                                                  help='set the default import action')
    set_import_action.set_defaults(func=config_set_import_action, write_to_history=False)
    set_import_action.add_argument('--action',
                                   help='the action to set',
                                   choices=['add', 'add_update_existing', 'update'])

    set_non_interactive = set_subparsers.add_parser('non-interactive',
                                                    help='set non-interactive mode to true')
    set_non_interactive.set_defaults(func=config_set_non_interactive, write_to_history=False)

    set_interactive = set_subparsers.add_parser('interactive',
                                                help='set non-interactive mode to false ')
    set_interactive.set_defaults(func=config_set_interactive, write_to_history=False)

    add = config_subparsers.add_parser('add',
                                       help='add values in the configuration file')
    add_subparsers = add.add_subparsers(metavar='')

    add_host = add_subparsers.add_parser('host',
                                         help='add a new host')
    add_host.set_defaults(func=config_add_host,
                          write_to_history=False)
    add_host.add_argument('--url',
                          help='the URL of the host')
    add_host.add_argument('--username',
                          help='the username of the admin account')
    add_host.add_argument('--password',
                          help='the password of the admin account')


# =======
# Methods
# =======

@command
def config_set_host(args):
    if args.url:
        url = args.url
    else:
        auths = config.get('host', 'auth')
        urls = [auth['url'] for auth in auths]
        url = ask.multi_choice('Please select a host:', urls)

    io.start("Switching to host {}".format(highlight(url)))

    if config.host_exists(url):
        config.set_host(url)
    else:
        raise McmdError("There is no host {} in the config file".format(url))


# noinspection PyUnusedLocal
@command
def config_set_import_action(args):
    if args.action:
        action = args.action
    else:
        options = ['add', 'add_update_existing', 'update']
        action = ask.multi_choice('Choose the default import action:', options)

    io.start("Setting import action to {}".format(highlight(action)))
    config.set_import_action(action)


# noinspection PyUnusedLocal
@command
def config_set_non_interactive(args):
    io.start('Switching to non-interactive mode')
    config.set_non_interactive(True)


# noinspection PyUnusedLocal
@command
def config_set_interactive(args):
    io.start('Switching to interactive mode')
    config.set_non_interactive(False)


# noinspection PyUnusedLocal
@command
def config_add_host(args):
    if args.url:
        url = args.url
    else:
        url = ask.input_("URL", required=True)

    if config.host_exists(url):
        raise McmdError("A host with URL {} already exists.".format(url))

    if args.username:
        username = args.username
    else:
        username = ask.input_("Username (Default: admin)")
        username = 'admin' if len(username) == 0 else username

    if args.password:
        password = args.password
    else:
        password = ask.password("Password (Leave blank to use command line authentication)")
        password = None if len(password) == 0 else password

    io.start("Adding host {}".format(highlight(url)))
    config.add_host(url, username, password)
    io.succeed()

    if not config.get('settings', 'non_interactive'):
        _switch_to_new_host(url)


def _switch_to_new_host(url):
    if ask.confirm("Do you want to switch to the new host?"):
        io.start("Switching to host {}".format(highlight(url)))
        config.set_host(url)



from setuptools import setup, find_packages

setup(
    name='molgenis-commander',
    version='1.10.1',
    packages=find_packages(),
    description='A command line interface for Molgenis',
    url='https://github.com/molgenis/molgenis-tools-commander',
    author='Tommy de Boer',
    author_email='tommydeboer4@gmail.com',
    license='GNU Lesser General Public License 3.0',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'mcmd = mcmd.__main__:main'
        ]
    },
    install_requires=[
        'attrs==20.3.0',
        'colorama==0.4.4',
        'halo==0.0.31',
        'Jinja2==2.11.3',
        'packaging==20.9',
        'parsy==1.3.0',
        'polling==0.3.1',
        'PyGithub==1.54.1',
        'questionary==1.9.0',
        'rainbow_logging_handler==2.2.2',
        'requests==2.21.0',
        'ruamel.yaml==0.17.2'
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'molgenis-py-client>=2.3.0',
        'pytest',
        'testfixtures'
    ]
)



import sys

# !!! DON'T PLACE IMPORT STATEMENTS BEFORE THE VERSION CHECK !!!
MIN_PYTHON = (3, 6)
if sys.version_info < MIN_PYTHON:
    import platform

    sys.exit(
        "Python {}.{} or later is required. You are running Python {}. Please upgrade to a newer version.\n".format(
            MIN_PYTHON[0],
            MIN_PYTHON[1],
            platform.python_version()))

import logging
import signal

from mcmd.core import update_checker, store
from mcmd.core.context.base_context import Context
from mcmd.args.errors import ArgumentSyntaxError
from mcmd.core.context.home_context import HomeContext
from mcmd.args.parser import parse_args
from mcmd.config.loader import load_config
from mcmd.io import io
from mcmd.io.io import set_debug
from mcmd.io.logging import set_level


def main():
    sys.exit(start(sys.argv, HomeContext()))


def start(argv, context: Context):
    with context:
        # setup friendly interrupt message
        signal.signal(signal.SIGINT, interrupt_handler)

        load_config()
        store.load()
        update_checker.check()

        args = _parse_args(argv)

        setattr(args, 'arg_string', ' '.join(argv[1:]))
        set_log_level(args)

        args.func(args)

        return 0


def _parse_args(argv):
    try:
        return parse_args(argv[1:])
    except ArgumentSyntaxError as e:
        sys.stderr.write(e.usage)
        sys.stderr.write(str(e))
        exit(1)


def set_log_level(args):
    if args.verbose:
        verbosity = int(args.verbose)
        if verbosity > 0:
            set_level(logging.DEBUG)
            set_debug()
    else:
        set_level(logging.INFO)


# noinspection PyUnusedLocal
def interrupt_handler(sig, frame):
    io.warn('Interrupted by user.')
    sys.exit(0)


if __name__ == '__main__':
    main()



"""
Give a principal (a user or a group) some permission on a resource (a package, entity type, plugin or row). Unless
specified by the user, the give command will try to figure out the principal- and resource types itself. If a resource
or principal doesn't exist, the program will terminate.
"""
import textwrap
from argparse import RawDescriptionHelpFormatter

from mcmd.commands._registry import arguments
from mcmd.core.command import command
from mcmd.core.compatibility import version
from mcmd.core.errors import McmdError
from mcmd.io import io
from mcmd.io.io import highlight
from mcmd.molgenis.principals import PrincipalType, get_principal_type_from_args
from mcmd.molgenis.resources import detect_resource_type, ensure_resource_exists, ResourceType
from mcmd.molgenis.security import security
from mcmd.molgenis.security.permission import Permission
from mcmd.molgenis.version import get_version


# =========
# Arguments
# =========

@arguments('give')
def add_arguments(subparsers):
    p_give = subparsers.add_parser('give',
                                   help='give permissions on resources to roles or users',
                                   formatter_class=RawDescriptionHelpFormatter,
                                   description=textwrap.dedent(
                                       """
                                       Give permissions on resources to roles or users. If you don't specify the 
                                       resource type (for example with --package or --plugin) or the receiver type 
                                       (--role or --user), the Commander will try to detect it on its own. 
                                       
                                       If you want to remove a permission, use the 'none' permission.

                                       Note: since MOLGENIS 8.3 role names are case sensitive and need to be typed 
                                       exactly as-is. (Before 8.3 all role names will be upper cased automatically). 

                                       example usage:
                                         mcmd give john read dataset
                                         mcmd give --user john read --entity-type dataset
                                         
                                         mcmd give group_EDITOR write dataexplorer
                                         mcmd give --role group_EDITOR write --plugin dataexplorer
                                         
                                         mcmd give john edit dataset --entity biobank1
                                         mcmd give --user john edit --entity-type dataset --entity biobank1
                                       """
                                   ))
    p_give.set_defaults(func=give,
                        write_to_history=True)
    p_give_resource = p_give.add_mutually_exclusive_group()
    p_give_resource.add_argument('--entity-type', '-e',
                                 action='store_true',
                                 help='flag to specify that the resource is an entity type (assumed when using '
                                      '--entity)')
    p_give_resource.add_argument('--package', '-p',
                                 action='store_true',
                                 help='flag to specify that the resource is a package')
    p_give_resource.add_argument('--plugin', '-pl',
                                 action='store_true',
                                 help='flag to specify that the resource is a plugin')
    p_give_receiver = p_give.add_mutually_exclusive_group()
    p_give.add_argument('--entity', '-E',
                        metavar='ENTITY_ID',
                        help='use in conjunction with --entity-type to give permissions for a row')
    p_give_receiver.add_argument('--user', '-u',
                                 action='store_true',
                                 help='flag to specify that the receiver is a user')
    p_give_receiver.add_argument('--role', '-r',
                                 action='store_true',
                                 help='flag to specify that the receiver is a role')
    p_give.add_argument('receiver',
                        type=str,
                        help='the role (or user) to give the permission to')
    p_give.add_argument('permission',
                        choices=['none', 'writemeta', 'readmeta', 'write', 'edit', 'read', 'view', 'count'],
                        help='the permission type to give - synonyms are allowed (e.g. write/edit)')
    p_give.add_argument('resource',
                        type=str,
                        help='the resource to which permission is given')


# =======
# Methods
# =======

@command
def give(args):
    _validate_args(args)

    permission = Permission[args.permission.upper()]

    principal_type = get_principal_type_from_args(args, principal_name=args.receiver)

    if args.entity:
        _grant_rls(principal_type=principal_type,
                   principal_name=args.receiver,
                   entity_type_id=args.resource,
                   entity_id=args.entity,
                   permission=permission)
    else:
        resource_type = _get_resource_type(args)
        _grant(principal_type=principal_type,
               principal_name=args.receiver,
               resource_type=resource_type,
               entity_type_id=args.resource,
               permission=permission)


@version('7.0.0')
def _validate_args(args):
    if args.entity:
        raise McmdError(
            "Giving row level permissions is only possible since MOLGENIS 8.1.1 (you are using {})".format(
                get_version()))


# noinspection PyUnusedLocal
@version('8.1.1')
def _validate_args(args):
    pass


def _grant(principal_type: PrincipalType, principal_name: str, resource_type: ResourceType, entity_type_id: str,
           permission: Permission):
    io.start('Giving %s %s permission to %s on %s %s' % (principal_type.value,
                                                         highlight(principal_name),
                                                         highlight(permission.value),
                                                         resource_type.get_label().lower(),
                                                         highlight(entity_type_id)))

    security.grant_permission(principal_type, principal_name, resource_type, entity_type_id, permission)


def _grant_rls(principal_type: PrincipalType, principal_name: str, entity_type_id: str,
               entity_id: str, permission: Permission):
    io.start('Giving %s %s permission to %s on row %s of entity type %s' % (principal_type.value,
                                                                            highlight(principal_name),
                                                                            highlight(permission.value),
                                                                            highlight(entity_id),
                                                                            highlight(entity_type_id)))

    security.grant_row_permission(principal_type, principal_name, entity_type_id, entity_id, permission)


def _get_resource_type(args):
    # The permission manager doesn't check if the resources actually exist so we need to do that ourselves

    resource_id = args.resource
    if args.entity_type:
        ensure_resource_exists(resource_id, ResourceType.ENTITY_TYPE)
        return ResourceType.ENTITY_TYPE
    elif args.package:
        ensure_resource_exists(resource_id, ResourceType.PACKAGE)
        return ResourceType.PACKAGE
    elif args.plugin:
        ensure_resource_exists(resource_id, ResourceType.PLUGIN)
        return ResourceType.PLUGIN
    else:
        return detect_resource_type(resource_id, [ResourceType.ENTITY_TYPE, ResourceType.PACKAGE, ResourceType.PLUGIN])



from colorama import Fore, Style
from halo import Halo

import mcmd.config.config as config
from mcmd.io.kbhit import KBHit
from mcmd.io.logging import get_logger

log = get_logger()

_debug_mode = False
spinner = None


def start(message):
    global spinner
    spinner = _new_spinner()
    spinner.start(message)


def succeed():
    global spinner
    if spinner:
        if config.has_option('settings', 'unicorn_mode') and config.get('settings', 'unicorn_mode'):
            spinner.stop_and_persist(symbol='🦄'.encode('utf-8'))
        else:
            spinner.succeed()

        spinner = None


def info(message):
    """Replaces an existing spinner with an info message and restarts the previous spinner afterwards. Just shows the
    info message if there was no spinner running."""
    prev_text = None
    if spinner:
        prev_text = spinner.text
        info_spinner = spinner
    else:
        info_spinner = _new_spinner()

    info_spinner.info(message)

    if prev_text:
        spinner.start(prev_text)


def warn(message):
    """Turns the current spinner in a warning message (with icon) and starts a new spinner"""
    if spinner:
        spinner.warn()
    log.warn('  ' + message)

    if spinner:
        spinner.start()


def pause():
    """Pause the spinner (with a warning icon), restart it with unpause()"""
    if spinner:
        spinner.warn()


def unpause():
    """Unpause a paused spinner"""
    if spinner:
        spinner.start()


def error(message):
    global spinner
    if spinner:
        spinner.fail()
        spinner = None

    if message:
        log.error('  ' + message.strip('\"\''))


def debug(message):
    global spinner
    if not _debug_mode:
        return

    if spinner:
        spinner.stop_and_persist()
        spinner = None

    log.debug('  ' + message)


def wait_for_enter():
    """Waits until the user presses enter. Non-blocking: the program can still be interrupted and closed."""
    kb = KBHit()
    while True:
        if kb.kbhit():
            c = kb.getch()
            if c in ['\n', '\r\n', '\r']:  # Enter
                break


def _new_spinner():
    return Halo(spinner='dots')


def set_debug():
    global _debug_mode
    _debug_mode = True


def newline():
    log.info('')


def highlight(string):
    return Fore.LIGHTBLUE_EX + string + Fore.RESET


def bold(string):
    return Style.BRIGHT + string + Style.RESET_ALL


def dim(string):
    return Style.DIM + string + Style.RESET_ALL
